import { __awaiter, __generator, __values, __spreadArray, __read } from 'tslib';
import { gqlPluckFromCodeString, gqlPluckFromCodeStringSync } from '@graphql-tools/graphql-tag-pluck/es5';
import micromatch from 'micromatch';
import unixify from 'unixify';
import { execFile, execFileSync } from 'child_process';
import os from 'os';
import { parseGraphQLSDL, parseGraphQLJSON, asArray } from '@graphql-tools/utils/es5';
import { parse as parse$1 } from 'graphql';
import isGlob from 'is-glob';

var createLoadError = function (error) { return new Error('Unable to load file from git: ' + error); };
var createShowCommand = function (_a) {
    var ref = _a.ref, path = _a.path;
    return ['show', ref + ":" + path];
};
var createTreeError = function (error) { return new Error('Unable to load the file tree from git: ' + error); };
var createTreeCommand = function (_a) {
    var ref = _a.ref;
    return ['ls-tree', '-r', '--name-only', ref];
};
/**
 * @internal
 */
function readTreeAtRef(ref) {
    return __awaiter(this, void 0, void 0, function () {
        var error_1;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    _a.trys.push([0, 2, , 3]);
                    return [4 /*yield*/, new Promise(function (resolve, reject) {
                            execFile('git', createTreeCommand({ ref: ref }), { encoding: 'utf-8', maxBuffer: 1024 * 1024 * 1024 }, function (error, stdout) {
                                if (error) {
                                    reject(error);
                                }
                                else {
                                    resolve(stdout.split(os.EOL).map(function (line) { return line.trim(); }));
                                }
                            });
                        })];
                case 1: return [2 /*return*/, _a.sent()];
                case 2:
                    error_1 = _a.sent();
                    throw createTreeError(error_1);
                case 3: return [2 /*return*/];
            }
        });
    });
}
/**
 * @internal
 */
function readTreeAtRefSync(ref) {
    try {
        return execFileSync('git', createTreeCommand({ ref: ref }), { encoding: 'utf-8' })
            .split(os.EOL)
            .map(function (line) { return line.trim(); });
    }
    catch (error) {
        throw createTreeError(error);
    }
}
/**
 * @internal
 */
function loadFromGit(input) {
    return __awaiter(this, void 0, void 0, function () {
        var error_2;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    _a.trys.push([0, 2, , 3]);
                    return [4 /*yield*/, new Promise(function (resolve, reject) {
                            execFile('git', createShowCommand(input), { encoding: 'utf-8', maxBuffer: 1024 * 1024 * 1024 }, function (error, stdout) {
                                if (error) {
                                    reject(error);
                                }
                                else {
                                    resolve(stdout);
                                }
                            });
                        })];
                case 1: return [2 /*return*/, _a.sent()];
                case 2:
                    error_2 = _a.sent();
                    throw createLoadError(error_2);
                case 3: return [2 /*return*/];
            }
        });
    });
}
/**
 * @internal
 */
function loadFromGitSync(input) {
    try {
        return execFileSync('git', createShowCommand(input), { encoding: 'utf-8' });
    }
    catch (error) {
        throw createLoadError(error);
    }
}

/**
 * @internal
 */
function parse(_a) {
    var path = _a.path, pointer = _a.pointer, content = _a.content, options = _a.options;
    if (/\.(gql|graphql)s?$/i.test(path)) {
        return parseGraphQLSDL(pointer, content, options);
    }
    if (/\.json$/i.test(path)) {
        return parseGraphQLJSON(pointer, content, options);
    }
}

// git:branch:path/to/file
function extractData(pointer) {
    var parts = pointer.replace(/^git\:/i, '').split(':');
    if (!parts || parts.length !== 2) {
        return null;
    }
    return {
        ref: parts[0],
        path: parts[1],
    };
}
/**
 * This loader loads a file from git.
 *
 * ```js
 * const typeDefs = await loadTypedefs('git:someBranch:some/path/to/file.js', {
 *   loaders: [new GitLoader()],
 * })
 * ```
 */
var GitLoader = /** @class */ (function () {
    function GitLoader() {
    }
    GitLoader.prototype.canLoad = function (pointer) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                return [2 /*return*/, this.canLoadSync(pointer)];
            });
        });
    };
    GitLoader.prototype.canLoadSync = function (pointer) {
        return typeof pointer === 'string' && pointer.toLowerCase().startsWith('git:');
    };
    GitLoader.prototype.resolveGlobs = function (glob, ignores) {
        return __awaiter(this, void 0, void 0, function () {
            var data, refsForPaths, ref, path, ignores_1, ignores_1_1, ignore, data_1, ref_1, path_1, resolved;
            var e_1, _a;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        data = extractData(glob);
                        if (data === null) {
                            return [2 /*return*/, []];
                        }
                        refsForPaths = new Map();
                        ref = data.ref, path = data.path;
                        if (!refsForPaths.has(ref)) {
                            refsForPaths.set(ref, []);
                        }
                        refsForPaths.get(ref).push(unixify(path));
                        try {
                            for (ignores_1 = __values(ignores), ignores_1_1 = ignores_1.next(); !ignores_1_1.done; ignores_1_1 = ignores_1.next()) {
                                ignore = ignores_1_1.value;
                                data_1 = extractData(ignore);
                                if (data_1 === null) {
                                    continue;
                                }
                                ref_1 = data_1.ref, path_1 = data_1.path;
                                if (!refsForPaths.has(ref_1)) {
                                    refsForPaths.set(ref_1, []);
                                }
                                refsForPaths.get(ref_1).push("!" + unixify(path_1));
                            }
                        }
                        catch (e_1_1) { e_1 = { error: e_1_1 }; }
                        finally {
                            try {
                                if (ignores_1_1 && !ignores_1_1.done && (_a = ignores_1.return)) _a.call(ignores_1);
                            }
                            finally { if (e_1) throw e_1.error; }
                        }
                        resolved = [];
                        return [4 /*yield*/, Promise.all(__spreadArray([], __read(refsForPaths.entries())).map(function (_a) {
                                var _b = __read(_a, 2), ref = _b[0], paths = _b[1];
                                return __awaiter(_this, void 0, void 0, function () {
                                    var _c, _d, _e, _f, _g;
                                    return __generator(this, function (_h) {
                                        switch (_h.label) {
                                            case 0:
                                                _d = (_c = resolved.push).apply;
                                                _e = [resolved];
                                                _f = [[]];
                                                _g = micromatch;
                                                return [4 /*yield*/, readTreeAtRef(ref)];
                                            case 1:
                                                _d.apply(_c, _e.concat([__spreadArray.apply(void 0, _f.concat([__read.apply(void 0, [_g.apply(void 0, [_h.sent(), paths]).map(function (filePath) { return "git:" + ref + ":" + filePath; })])]))]));
                                                return [2 /*return*/];
                                        }
                                    });
                                });
                            }))];
                    case 1:
                        _b.sent();
                        return [2 /*return*/, resolved];
                }
            });
        });
    };
    GitLoader.prototype.resolveGlobsSync = function (glob, ignores) {
        var e_2, _a, e_3, _b;
        var data = extractData(glob);
        if (data === null) {
            return [];
        }
        var ref = data.ref, path = data.path;
        var refsForPaths = new Map();
        if (!refsForPaths.has(ref)) {
            refsForPaths.set(ref, []);
        }
        refsForPaths.get(ref).push(unixify(path));
        try {
            for (var ignores_2 = __values(ignores), ignores_2_1 = ignores_2.next(); !ignores_2_1.done; ignores_2_1 = ignores_2.next()) {
                var ignore = ignores_2_1.value;
                var data_2 = extractData(ignore);
                if (data_2 === null) {
                    continue;
                }
                var ref_2 = data_2.ref, path_2 = data_2.path;
                if (!refsForPaths.has(ref_2)) {
                    refsForPaths.set(ref_2, []);
                }
                refsForPaths.get(ref_2).push("!" + unixify(path_2));
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (ignores_2_1 && !ignores_2_1.done && (_a = ignores_2.return)) _a.call(ignores_2);
            }
            finally { if (e_2) throw e_2.error; }
        }
        var resolved = [];
        var _loop_1 = function (ref_3, paths) {
            resolved.push.apply(resolved, __spreadArray([], __read(micromatch(readTreeAtRefSync(ref_3), paths).map(function (filePath) { return "git:" + ref_3 + ":" + filePath; }))));
        };
        try {
            for (var _c = __values(refsForPaths.entries()), _d = _c.next(); !_d.done; _d = _c.next()) {
                var _e = __read(_d.value, 2), ref_3 = _e[0], paths = _e[1];
                _loop_1(ref_3, paths);
            }
        }
        catch (e_3_1) { e_3 = { error: e_3_1 }; }
        finally {
            try {
                if (_d && !_d.done && (_b = _c.return)) _b.call(_c);
            }
            finally { if (e_3) throw e_3.error; }
        }
        return resolved;
    };
    GitLoader.prototype._load = function (pointer, options) {
        return __awaiter(this, void 0, void 0, function () {
            var result, ref, path, content, parsed, sources;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        result = extractData(pointer);
                        if (result === null) {
                            return [2 /*return*/, []];
                        }
                        ref = result.ref, path = result.path;
                        return [4 /*yield*/, loadFromGit({ ref: ref, path: path })];
                    case 1:
                        content = _a.sent();
                        parsed = parse({ path: path, options: options, pointer: pointer, content: content });
                        if (parsed) {
                            return [2 /*return*/, [parsed]];
                        }
                        return [4 /*yield*/, gqlPluckFromCodeString(pointer, content, options.pluckConfig)];
                    case 2:
                        sources = _a.sent();
                        return [2 /*return*/, sources.map(function (source) { return ({
                                location: pointer,
                                document: parse$1(source, options),
                            }); })];
                }
            });
        });
    };
    GitLoader.prototype.load = function (pointer, options) {
        return __awaiter(this, void 0, void 0, function () {
            var result, path, finalResult, resolvedPaths, _a, _b, _c, _d;
            var _this = this;
            return __generator(this, function (_e) {
                switch (_e.label) {
                    case 0:
                        result = extractData(pointer);
                        if (result === null) {
                            return [2 /*return*/, []];
                        }
                        path = result.path;
                        finalResult = [];
                        if (!isGlob(path)) return [3 /*break*/, 3];
                        return [4 /*yield*/, this.resolveGlobs(pointer, asArray(options.ignore || []))];
                    case 1:
                        resolvedPaths = _e.sent();
                        return [4 /*yield*/, Promise.all(resolvedPaths.map(function (path) { return __awaiter(_this, void 0, void 0, function () {
                                var results;
                                return __generator(this, function (_a) {
                                    switch (_a.label) {
                                        case 0: return [4 /*yield*/, this.load(path, options)];
                                        case 1:
                                            results = _a.sent();
                                            finalResult.push.apply(finalResult, __spreadArray([], __read(results)));
                                            return [2 /*return*/];
                                    }
                                });
                            }); }))];
                    case 2:
                        _e.sent();
                        return [3 /*break*/, 6];
                    case 3: return [4 /*yield*/, this.canLoad(pointer)];
                    case 4:
                        if (!_e.sent()) return [3 /*break*/, 6];
                        _b = (_a = finalResult.push).apply;
                        _c = [finalResult];
                        _d = [[]];
                        return [4 /*yield*/, this._load(pointer, options)];
                    case 5:
                        _b.apply(_a, _c.concat([__spreadArray.apply(void 0, _d.concat([__read.apply(void 0, [(_e.sent())])]))]));
                        _e.label = 6;
                    case 6: return [2 /*return*/, finalResult];
                }
            });
        });
    };
    GitLoader.prototype._loadSync = function (pointer, options) {
        var result = extractData(pointer);
        if (result === null) {
            return [];
        }
        var ref = result.ref, path = result.path;
        var content = loadFromGitSync({ ref: ref, path: path });
        var parsed = parse({ path: path, options: options, pointer: pointer, content: content });
        if (parsed) {
            return [parsed];
        }
        var sources = gqlPluckFromCodeStringSync(pointer, content, options.pluckConfig);
        return sources.map(function (source) { return ({
            location: pointer,
            document: parse$1(source, options),
        }); });
    };
    GitLoader.prototype.loadSync = function (pointer, options) {
        var e_4, _a;
        var result = extractData(pointer);
        if (result === null) {
            return [];
        }
        var path = result.path;
        var finalResult = [];
        if (isGlob(path)) {
            var resolvedPaths = this.resolveGlobsSync(pointer, asArray(options.ignore || []));
            var finalResult_1 = [];
            try {
                for (var resolvedPaths_1 = __values(resolvedPaths), resolvedPaths_1_1 = resolvedPaths_1.next(); !resolvedPaths_1_1.done; resolvedPaths_1_1 = resolvedPaths_1.next()) {
                    var path_3 = resolvedPaths_1_1.value;
                    if (this.canLoadSync(path_3)) {
                        finalResult_1.push.apply(finalResult_1, __spreadArray([], __read(this.loadSync(path_3, options))));
                    }
                }
            }
            catch (e_4_1) { e_4 = { error: e_4_1 }; }
            finally {
                try {
                    if (resolvedPaths_1_1 && !resolvedPaths_1_1.done && (_a = resolvedPaths_1.return)) _a.call(resolvedPaths_1);
                }
                finally { if (e_4) throw e_4.error; }
            }
        }
        else if (this.canLoadSync(pointer)) {
            finalResult.push.apply(finalResult, __spreadArray([], __read(this._loadSync(pointer, options))));
        }
        return finalResult;
    };
    return GitLoader;
}());

export { GitLoader };
